@@ -1,22 +1,23 @@
 /* key.c - Key objects.
    Copyright (C) 2000 Werner Koch (dd9jn)
-   Copyright (C) 2001, 2002, 2003 g10 Code GmbH
+   Copyright (C) 2001, 2002, 2003, 2004 g10 Code GmbH
 
    This file is part of GPGME.
  
    GPGME is free software; you can redistribute it and/or modify it
-   under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
- 
+   under the terms of the GNU Lesser General Public License as
+   published by the Free Software Foundation; either version 2.1 of
+   the License, or (at your option) any later version.
+   
    GPGME is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   General Public License for more details.
- 
-   You should have received a copy of the GNU General Public License
-   along with GPGME; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Lesser General Public License for more details.
+   
+   You should have received a copy of the GNU Lesser General Public
+   License along with this program; if not, write to the Free Software
+   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
+   02111-1307, USA.  */
 
 #if HAVE_CONFIG_H
 #include <config.h>
@@ -220,7 +221,7 @@ _gpgme_key_append_name (gpgme_key_t key, char *src)
   dst = uid->uid;
   _gpgme_decode_c_string (src, &dst, src_len + 1);
 
-  dst += src_len + 1;
+  dst += strlen (dst) + 1;
   if (key->protocol == GPGME_PROTOCOL_CMS)
     parse_x509_user_id (uid->uid, &uid->name, &uid->email,
 			&uid->comment, dst);
@@ -253,9 +254,11 @@ _gpgme_key_add_sig (gpgme_key_t key, char *src)
   /* We can malloc a buffer of the same length, because the converted
      string will never be larger. Actually we allocate it twice the
      size, so that we are able to store the parsed stuff there too.  */
-  sig = calloc (1, sizeof (*sig) + 2 * src_len + 3);
+  sig = malloc (sizeof (*sig) + 2 * src_len + 3);
   if (!sig)
     return NULL;
+  memset (sig, 0, sizeof *sig);
+
   sig->keyid = sig->_keyid;
   sig->_keyid[16] = '\0';
   sig->uid = ((char *) sig) + sizeof (*sig);
@@ -264,7 +267,7 @@ _gpgme_key_add_sig (gpgme_key_t key, char *src)
     {
       char *dst = sig->uid;
       _gpgme_decode_c_string (src, &dst, src_len + 1);
-      dst += src_len + 1;
+      dst += strlen (dst) + 1;
       if (key->protocol == GPGME_PROTOCOL_CMS)
 	parse_x509_user_id (sig->uid, &sig->name, &sig->email,
 			    &sig->comment, dst);
@@ -304,6 +307,9 @@ gpgme_key_unref (gpgme_key_t key)
   gpgme_user_id_t uid;
   gpgme_subkey_t subkey;
 
+  if (!key)
+    return;
+
   LOCK (key_ref_lock);
   assert (key->_refs > 0);
   if (--key->_refs)
@@ -331,9 +337,19 @@ gpgme_key_unref (gpgme_key_t key)
 
       while (keysig)
 	{
-	  gpgme_key_sig_t next = keysig->next;
+	  gpgme_key_sig_t next_keysig = keysig->next;
+	  gpgme_sig_notation_t notation = keysig->notations;
+
+	  while (notation)
+	    {
+	      gpgme_sig_notation_t next_notation = notation->next;
+
+	      _gpgme_sig_notation_free (notation);
+	      notation = next_notation;
+	    }
+
           free (keysig);
-	  keysig = next;
+	  keysig = next_keysig;
         }
       free (uid);
       uid = next_uid;
@@ -695,7 +711,7 @@ gpgme_key_sig_get_ulong_attr (gpgme_key_t key, int uid_idx, _gpgme_attr_t what,
       return certsig->expired;
 
     case GPGME_ATTR_SIG_CLASS:
-      return certsig->class;
+      return certsig->sig_class;
 
     case GPGME_ATTR_SIG_STATUS:
       return certsig->status;
